@@ -7,6 +7,7 @@ module Zeitwerk
   class Loader
     require_relative "loader/callbacks"
     include Callbacks
+    include RealModName
 
     # @return [String]
     attr_reader :tag
@@ -43,7 +44,7 @@ module Zeitwerk
     #
     # @private
     # @return [Set<String>]
-    attr_reader :ignored
+    attr_reader :ignored_glob_patterns
 
     # The actual collection of absolute file and directory names at the time the
     # ignored glob patterns were expanded. Computed on setup, and recomputed on
@@ -53,23 +54,6 @@ module Zeitwerk
     # @return [Set<String>]
     attr_reader :ignored_paths
 
-    # A _shadowed file_ is a file managed by this loader that is ignored when
-    # setting autoloads because its matching constant is taken. Either the
-    # constant is already defined, or there exists an autoload for it.
-    #
-    # Think $LOAD_PATH and require, only the first occurrence of a given
-    # relative name is loaded.
-    #
-    # This set keeps track of the absolute path of shadowed files, to be able to
-    # skip them while eager loading.
-    #
-    # Note this cannot be implemented with do_not_eager_load because these
-    # files are not autoloadable.
-    #
-    # @private
-    # @return [Set<String>]
-    attr_reader :shadowed_files
-
     # Maps real absolute paths for which an autoload has been set ---and not
     # executed--- to their corresponding parent class or module and constant
     # name.
@@ -149,13 +133,12 @@ module Zeitwerk
 
       @root_dirs             = {}
       @preloads              = []
-      @ignored               = Set.new
+      @ignored_glob_patterns = Set.new
       @ignored_paths         = Set.new
       @autoloads             = {}
       @autoloaded_dirs       = []
       @to_unload             = {}
       @lazy_subdirs          = {}
-      @shadowed_files        = Set.new
       @eager_load_exclusions = Set.new
 
       # TODO: find a better name for these mutexes.
@@ -242,16 +225,12 @@ module Zeitwerk
     #
     # @param paths [<String, Pathname, <String, Pathname>>]
     # @return [void]
-    def ignore(*paths)
-      mutex.synchronize { ignored.merge(expand_paths(paths)) }
-    end
-
-    # @private
-    # @return [void]
-    def expand_ignored_glob_patterns
-      # Note that Dir.glob works with regular file names just fine. That is,
-      # glob patterns technically need no wildcards.
-      ignored_paths.replace(ignored.flat_map { |path| Dir.glob(path) })
+    def ignore(*glob_patterns)
+      glob_patterns = expand_paths(glob_patterns)
+      mutex.synchronize do
+        ignored_glob_patterns.merge(glob_patterns)
+        ignored_paths.merge(expand_glob_patterns(glob_patterns))
+      end
     end
 
     # Sets autoloads in the root namespace and preloads files, if any.
@@ -261,7 +240,6 @@ module Zeitwerk
       mutex.synchronize do
         break if @setup
 
-        expand_ignored_glob_patterns
         actual_root_dirs.each { |root_dir| set_autoloads_in_dir(root_dir, Object) }
         do_preload
 
@@ -324,7 +302,6 @@ module Zeitwerk
         autoloaded_dirs.clear
         to_unload.clear
         lazy_subdirs.clear
-        shadowed_files.clear
 
         Registry.on_unload(self)
         ExplicitNamespace.unregister(self)
@@ -344,6 +321,7 @@ module Zeitwerk
     def reload
       if reloading_enabled?
         unload
+        recompute_ignored_paths
         setup
       else
         raise ReloadingDisabledError, "can't reload, please call loader.enable_reloading before setup"
@@ -361,20 +339,24 @@ module Zeitwerk
         break if @eager_loaded
 
         queue = actual_root_dirs.reject { |dir| eager_load_exclusions.member?(dir) }
-        while dir = queue.shift
-          ls(dir) do |_basename, abspath|
+        queue.map! { |dir| [Object, dir] }
+        while to_eager_load = queue.shift
+          namespace, dir = to_eager_load
+
+          ls(dir) do |basename, abspath|
             next if eager_load_exclusions.member?(abspath)
 
             if ruby?(abspath)
-              require abspath unless shadowed_files.member?(abspath)
-            elsif dir?(abspath)
-              queue << abspath
+              if cref = autoloads[File.realpath(abspath)]
+                cref[0].const_get(cref[1], false)
+              end
+            elsif dir?(abspath) && !root_dirs.key?(abspath)
+              cname = inflector.camelize(basename, abspath)
+              queue << [namespace.const_get(cname, false), abspath]
             end
           end
         end
 
-        shadowed_files.clear
-
         autoloaded_dirs.each do |autoloaded_dir|
           Registry.unregister_autoload(autoloaded_dir)
         end
@@ -417,6 +399,22 @@ module Zeitwerk
       @logger = ->(msg) { puts msg }
     end
 
+    # @private
+    # @param dir [String]
+    # @return [Boolean]
+    def manages?(dir)
+      dir = dir + "/"
+      ignored_paths.each do |ignored_path|
+        return false if dir.start_with?(ignored_path + "/")
+      end
+
+      root_dirs.each_key do |root_dir|
+        return true if root_dir.start_with?(dir) || dir.start_with?(root_dir + "/")
+      end
+
+      false
+    end
+
     # --- Class methods ---------------------------------------------------------------------------
 
     class << self
@@ -476,21 +474,38 @@ module Zeitwerk
     # @return [void]
     def set_autoloads_in_dir(dir, parent)
       ls(dir) do |basename, abspath|
-        if ruby?(basename)
-          basename.slice!(-3, 3)
-          cname = inflector.camelize(basename, abspath).to_sym
-          autoload_file(parent, cname, abspath)
-        elsif dir?(abspath)
-          # In a Rails application, `app/models/concerns` is a subdirectory of
-          # `app/models`, but both of them are root directories.
-          #
-          # To resolve the ambiguity file name -> constant path this introduces,
-          # the `app/models/concerns` directory is totally ignored as a namespace,
-          # it counts only as root. The guard checks that.
-          unless root_dirs.key?(abspath)
+        begin
+          if ruby?(basename)
+            basename.slice!(-3, 3)
             cname = inflector.camelize(basename, abspath).to_sym
-            autoload_subdir(parent, cname, abspath)
+            autoload_file(parent, cname, abspath)
+          elsif dir?(abspath)
+            # In a Rails application, `app/models/concerns` is a subdirectory of
+            # `app/models`, but both of them are root directories.
+            #
+            # To resolve the ambiguity file name -> constant path this introduces,
+            # the `app/models/concerns` directory is totally ignored as a namespace,
+            # it counts only as root. The guard checks that.
+            unless root_dirs.key?(abspath)
+              cname = inflector.camelize(basename, abspath).to_sym
+              autoload_subdir(parent, cname, abspath)
+            end
           end
+        rescue ::NameError => error
+          path_type = ruby?(abspath) ? "file" : "directory"
+
+          raise NameError.new(<<~MESSAGE, error.name)
+            #{error.message} inferred by #{inflector.class} from #{path_type}
+
+              #{abspath}
+
+            Possible ways to address this:
+
+              * Tell Zeitwerk to ignore this particular #{path_type}.
+              * Tell Zeitwerk to ignore one of its parent directories.
+              * Rename the #{path_type} to comply with the naming conventions.
+              * Modify the inflector to handle this case.
+          MESSAGE
         end
       end
     end
@@ -526,7 +541,6 @@ module Zeitwerk
       if autoload_path = autoload_for?(parent, cname)
         # First autoload for a Ruby file wins, just ignore subsequent ones.
         if ruby?(autoload_path)
-          shadowed_files.add(file)
           log("file #{file} is ignored because #{autoload_path} has precedence") if logger
         else
           promote_namespace_from_implicit_to_explicit(
@@ -537,7 +551,6 @@ module Zeitwerk
           )
         end
       elsif cdef?(parent, cname)
-        shadowed_files.add(file)
         log("file #{file} is ignored because #{cpath(parent, cname)} is already defined") if logger
       else
         set_autoload(parent, cname, file)
@@ -545,7 +558,7 @@ module Zeitwerk
     end
 
     # @param dir [String] directory that would have autovivified a module
-    # @param file [String] the file where the namespace is explictly defined
+    # @param file [String] the file where the namespace is explicitly defined
     # @param parent [Module]
     # @param cname [Symbol]
     # @return [void]
@@ -610,8 +623,14 @@ module Zeitwerk
     # @param parent [Module]
     # @param cname [Symbol]
     # @return [String, nil]
-    def strict_autoload_path(parent, cname)
-      parent.autoload?(cname) if cdef?(parent, cname)
+    if method(:autoload?).arity == 1
+      def strict_autoload_path(parent, cname)
+        parent.autoload?(cname) if cdef?(parent, cname)
+      end
+    else
+      def strict_autoload_path(parent, cname)
+        parent.autoload?(cname, false)
+      end
     end
 
     # This method is called this way because I prefer `preload` to be the method
@@ -653,7 +672,7 @@ module Zeitwerk
     # @param cname [Symbol]
     # @return [String]
     def cpath(parent, cname)
-      parent.equal?(Object) ? cname.to_s : "#{parent.name}::#{cname}"
+      parent.equal?(Object) ? cname.to_s : "#{real_mod_name(parent)}::#{cname}"
     end
 
     # @param dir [String]
@@ -682,7 +701,20 @@ module Zeitwerk
     # @param paths [<String, Pathname, <String, Pathname>>]
     # @return [<String>]
     def expand_paths(paths)
-      Array(paths).flatten.map! { |path| File.expand_path(path) }
+      paths.flatten.map! { |path| File.expand_path(path) }
+    end
+
+    # @param glob_patterns [<String>]
+    # @return [<String>]
+    def expand_glob_patterns(glob_patterns)
+      # Note that Dir.glob works with regular file names just fine. That is,
+      # glob patterns technically need no wildcards.
+      glob_patterns.flat_map { |glob_pattern| Dir.glob(glob_pattern) }
+    end
+
+    # @return [void]
+    def recompute_ignored_paths
+      ignored_paths.replace(expand_glob_patterns(ignored_glob_patterns))
     end
 
     # @param message [String]
@@ -703,16 +735,12 @@ module Zeitwerk
     def raise_if_conflicting_directory(dir)
       self.class.mutex.synchronize do
         Registry.loaders.each do |loader|
-          next if loader == self
-
-          loader.dirs.each do |already_managed_dir|
-            if dir.start_with?(already_managed_dir) || already_managed_dir.start_with?(dir)
-              require "pp"
-              raise Error,
-                "loader\n\n#{pretty_inspect}\n\nwants to manage directory #{dir}," \
-                " which is already managed by\n\n#{loader.pretty_inspect}\n"
-              EOS
-            end
+          if loader != self && loader.manages?(dir)
+            require "pp"
+            raise Error,
+              "loader\n\n#{pretty_inspect}\n\nwants to manage directory #{dir}," \
+              " which is already managed by\n\n#{loader.pretty_inspect}\n"
+            EOS
           end
         end
       end
